from flask import abort, flash, Markup, redirect, render_template, url_for
from flask_login import login_required, current_user
from app import db
from app.models import SpaCyNLPPipelineModel
from . import bp, TEMPLATE_FOLDER
from .forms import (
    CreateSpaCyNLPPipelineModelForm,
    EditSpaCyNLPPipelineModelForm
)


@bp.route('')
@login_required
def spacy_nlp_pipeline_models():
    return render_template(
        f'{TEMPLATE_FOLDER}/spacy_nlp_pipeline_models.html.j2',
        title='SpaCy NLP Pipeline Models'
    )


@bp.route('/create', methods=['GET', 'POST'])
@login_required
def create_spacy_nlp_pipeline_model():
    form = CreateSpaCyNLPPipelineModelForm(prefix='create-spacy-nlp-pipeline-model-form')
    if form.is_submitted():
        if not form.validate():
            response = {'errors': form.errors}
            return response, 400
        try:
            spacy_nlp_pipeline_model = SpaCyNLPPipelineModel.create(
                form.spacy_model_file.data,
                compatible_service_versions=form.compatible_service_versions.data,
                description=form.description.data,
                pipeline_name=form.pipeline_name.data,
                publisher=form.publisher.data,
                publisher_url=form.publisher_url.data,
                publishing_url=form.publishing_url.data,
                publishing_year=form.publishing_year.data,
                is_public=False,
                title=form.title.data,
                version=form.version.data,
                user=current_user
            )
        except OSError:
            abort(500)
        db.session.commit()
        spacy_nlp_pipeline_model_url = url_for(
            '.spacy_nlp_pipeline_model',
            spacy_nlp_pipeline_model_id=spacy_nlp_pipeline_model.id
        )
        message = Markup(
            f'SpaCy NLP Pipeline model "{spacy_nlp_pipeline_model.title}" '
            'created'
        )
        flash(message)
        return '', 201, {'Location': spacy_nlp_pipeline_model_url}
    return render_template(
        f'{TEMPLATE_FOLDER}/create_spacy_nlp_pipeline_model.html.j2',
        form=form,
        title='Create SpaCy NLP Pipeline Model'
    )


@bp.route('/<hashid:spacy_nlp_pipeline_model_id>', methods=['GET', 'POST'])
@login_required
def spacy_nlp_pipeline_model(spacy_nlp_pipeline_model_id):
    spacy_nlp_pipeline_model = SpaCyNLPPipelineModel.query.get_or_404(spacy_nlp_pipeline_model_id)
    form = EditSpaCyNLPPipelineModelForm(
        data=spacy_nlp_pipeline_model.to_json_serializeable(),
        prefix='edit-spacy-nlp-pipeline-model-form'
    )
    if form.validate_on_submit():
        form.populate_obj(spacy_nlp_pipeline_model)
        if db.session.is_modified(spacy_nlp_pipeline_model):
            message = Markup(
                f'SpaCy NLP Pipeline model "{spacy_nlp_pipeline_model.title}" '
                'updated'
            )
            flash(message)
            db.session.commit()
        return redirect(url_for('.spacy_nlp_pipeline_models'))
    return render_template(
        f'{TEMPLATE_FOLDER}/spacy_nlp_pipeline_model.html.j2',
        form=form,
        spacy_nlp_pipeline_model=spacy_nlp_pipeline_model,
        title=f'{spacy_nlp_pipeline_model.title} {spacy_nlp_pipeline_model.version}'
    )
